import uvicorn
from django.conf import settings
from django.utils.module_loading import import_string
from django.core.management.commands import runserver

from ._base import BaseCommand


class Command(BaseCommand):
    help = runserver.Command.help
    default_addrport = settings.WEB_ADDRPORT

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--addrport', '-p',
            default=self.default_addrport,
            dest='addrport', help='Specifies the uwsgi address:port. Default: [:8080]',
        )
        parser.add_argument(
            '--no-access-log',
            action='store_false', dest='access_log', default=True,
            help="Disable access logs.",
        )

    def handle(self, *args, **opts):
        super().handle(*args, **opts)

        host, port = opts.pop('addrport').split(':')
        if not host:
            host = '127.0.0.1'

        uvicorn.run(
            app=import_string(settings.ASGI_APPLICATION),
            access_log=opts['access_log'],
            log_level=self.LOG_LEVEL.lower(),
            interface='asgi3',
            host=host,
            port=int(port),
        )



import typing as _t

from django.http.response import FileResponse
from rest_framework import serializers, viewsets
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.decorators import action
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList
from drf_yasg.utils import swagger_auto_schema

from .responses import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from .serializers import EmptySerializer, DataSerializer


class Action:
    """
    Base class of actions. Has minimal of required functionality to create an action and write only business logic.
    This decorator is suitable in cases where it is not possible to implement the logic using :class:`.SimpleAction`
    or the algorithm is much more complicated than standard CRUD.

    Examples:

       .. sourcecode:: python

        ...
        from vstutils.api.fields import VSTCharField
        from vstutils.api.serializers import BaseSerializer
        from vstutils.api.base import ModelViewSet
        from vstutils.api.actions import Action
        ...

        class RequestSerializer(BaseSerializer):
            data_field1 = ...
            ...


        class ResponseSerializer(BaseSerializer):
            detail = VSTCharField(read_only=True)


        class AuthorViewSet(ModelViewSet):
            model = ...
            ...

            @Action(serializer_class=RequestSerializer, result_serializer_class=ResponseSerializer, ...)
            def profile(self, request, *args, **kwargs):
                ''' Got `serializer_class` body and response with `result_serializer_class`. '''
                serializer = self.get_serializer(self.get_object(), data=request.data)
                serializer.is_valid(raise_exception=True)
                return {"detail": "Executed!"}



    :param detail: Flag indicating which type of action is used: on a list or on a single entity.
                   Affects where this action will be displayed - on a detailed record or on a list of records.
    :param methods: List of available HTTP-methods for this action. Default has only `POST` method.
    :param serializer_class: Request body serializer. Also used for default response.
    :param result_serializer_class: Response body serializer. Required, when request and response
                                    has different set of fields.
    :param query_serializer: GET-request query data serializer. It is used when it is necessary to get valid data
                             in the query data of a GET-request and cast it to the required type.
    :param multi: Used only with non-GET requests and notify GUI, that this action
                  should be rendered over the selected list items.
    :param title: Title for action in UI. For non-GET actions, title is generated from method's name.
    :param icons: List of icons for UI button.
    :param is_list: Flag indicating whether the action type is a list or a single entity.
                    Typically used with GET actions.
    :param kwargs: Set of named arguments for :func:`rest_framework.decorators.action`.

    """
    __slots__ = (
        'detail',
        'methods',
        'serializer_class',
        'result_serializer_class',
        'query_serializer',
        'action_kwargs',
        'multi',
        'title',
        'icons',
        'is_list',
    )
    method_response_mapping = {
        "HEAD": HTTP_200_OK,
        "GET": HTTP_200_OK,
        "PUT": HTTP_200_OK,
        "PATCH": HTTP_200_OK,
        "POST": HTTP_201_CREATED,
        "DELETE": HTTP_204_NO_CONTENT,
    }

    def __init__(  # noqa: CFQ002
        self,
        detail: bool = True,
        methods: _t.List[_t.Text] = None,
        serializer_class: _t.Type[serializers.Serializer] = DataSerializer,
        result_serializer_class: _t.Type[serializers.Serializer] = None,
        query_serializer: _t.Type[serializers.Serializer] = None,
        multi: bool = False,
        title: _t.Text = None,
        icons: _t.Union[_t.Text, _t.Iterable] = None,
        is_list: bool = False,
        **kwargs,
    ):
        # pylint: disable=too-many-arguments
        self.detail = detail
        self.methods = list(map(str.upper, methods or ['POST'])) or None
        self.serializer_class = serializer_class
        self.result_serializer_class = result_serializer_class
        self.query_serializer = query_serializer
        self.multi = multi
        self.title = title
        self.icons = icons
        self.is_list = is_list
        self.action_kwargs = kwargs
        if self.query_serializer:
            self.action_kwargs['query_serializer'] = self.query_serializer

    @property
    def is_page(self):
        return 'GET' in self.methods

    def wrap_function(self, func):
        res = action(
            detail=self.detail,
            serializer_class=self.serializer_class,
            methods=self.methods,
            **self.action_kwargs
        )(func)

        name = func.__name__.replace('_', ' ').capitalize()
        swagger_kwargs = {
            'operation_description': (func.__doc__ or name).strip(),
            'methods': self.methods,
            'x-multiaction': self.multi if not self.is_page else False,
            'x-title': self.title or (None if self.is_page else name),
        }
        if self.query_serializer:
            swagger_kwargs['query_serializer'] = self.query_serializer
        if self.icons:
            swagger_kwargs['x-icons'] = self.icons.split(' ') if isinstance(self.icons, _t.Text) else list(self.icons)
        if self.is_page:
            swagger_kwargs['x-list'] = self.is_list
        if self.result_serializer_class and issubclass(self.result_serializer_class, serializers.Serializer):
            swagger_kwargs['responses'] = {
                self.method_response_mapping[method].status_code: self.result_serializer_class
                for method in self.methods
            }

        res = swagger_auto_schema(**swagger_kwargs)(res)
        res.action = self
        res.__doc__ = swagger_kwargs['operation_description']
        return res

    def __call__(self, method):

        def action_method(
                view: viewsets.GenericViewSet,
                request: Request,
                *args,
                **kwargs,
        ) -> _t.Union[Response, FileResponse]:

            result_serializer_class: _t.Type[serializers.Serializer] = self.result_serializer_class  # type: ignore
            identical = False
            if result_serializer_class is None:
                result_serializer_class = view.get_serializer_class()  # type: ignore
                identical = True

            result = method(view, request, *args, **kwargs)
            response_class = self.method_response_mapping[_t.cast(_t.Text, request.method)]

            if issubclass(result_serializer_class, serializers.Serializer):
                if not (isinstance(result, (ReturnDict, ReturnList)) and identical):
                    serializer = result_serializer_class(
                        result,
                        many=self.is_list,
                        context=view.get_serializer_context()
                    )
                    result = serializer.data
            elif isinstance(result, FileResponse):
                return result

            if self.is_list and result:
                result = {
                    'count': len(result),
                    'results': result,
                }
            return response_class(result)

        action_method.__name__ = self.action_kwargs.get('name', method.__name__)
        action_method.__doc__ = method.__doc__
        return self.wrap_function(action_method)


class EmptyAction(Action):
    """
    In this case, actions on an object do not require any data and manipulations with them.
    For such cases, there is a standard method that allows you to simplify
    the scheme and code to work just with the object.

    Optionally, you can also overload the response serializer
    to notify the interface about the format of the returned data.


    Examples:

       .. sourcecode:: python

        ...
        from vstutils.api.fields import RedirectIntegerField
        from vstutils.api.serializers import BaseSerializer
        from vstutils.api.base import ModelViewSet
        from vstutils.api.actions import EmptyAction
        ...

        class ResponseSerializer(BaseSerializer):
            id = RedirectIntegerField(operation_name='sync_history')


        class AuthorViewSet(ModelViewSet):
            model = ...
            ...

            @EmptyAction(result_serializer_class=ResponseSerializer, ...)
            def sync_data(self, request, *args, **kwargs):
                ''' Example of action which get object, sync data and redirect user to another view. '''
                sync_id = self.get_object().sync().id
                return {"id": sync_id}


       .. sourcecode:: python

        ...
        from django.http.response import FileResponse
        from vstutils.api.base import ModelViewSet
        from vstutils.api.actions import EmptyAction
        ...

        class AuthorViewSet(ModelViewSet):
            model = ...
            ...

            @EmptyAction(result_serializer_class=ResponseSerializer, ...)
            def resume(self, request, *args, **kwargs):
                ''' Example of action which response with generated resume in pdf. '''
                instance = self.get_object()

                return FileResponse(
                    streaming_content=instance.get_pdf(),
                    as_attachment=True,
                    filename=f'{instance.last_name}_{instance.first_name}_resume.pdf'
                )

    """

    __slots__ = ()

    def __init__(self, **kwargs):
        kwargs['serializer_class'] = EmptySerializer
        super().__init__(**kwargs)


class SimpleAction(Action):
    """
    The idea of this decorator is to get the full CRUD for the instance in a minimum of steps.
    The instance is the object that was returned from the method being decorated.
    The whole mechanism is very similar to the standard property decorator,
    with a description of a getter, setter, and deleter.

    If you're going to create an entry point for working with a single object, then you do not need to define methods.
    The presence of a getter, setter, and deleter will determine which methods will be available.

    In the official documentation of Django, an example is given with moving data
    that is not important for authorization to the Profile model.
    To work with such data that is outside the main model, there is this action object,
    which implements the basic logic in the most automated way.

    It covers most of the tasks for working with such data. By default, it has a GET method instead of POST.
    Also, for better organization of the code, it allows you to change the methods
    that will be called when modifying or deleting data.

    When assigning an action on an object, the list of methods is also filled with the necessary ones.

    Examples:

       .. sourcecode:: python

        ...
        from vstutils.api.fields import PhoneField
        from vstutils.api.serializers import BaseSerializer
        from vstutils.api.base import ModelViewSet
        from vstutils.api.actions import Action
        ...

        class ProfileSerializer(BaseSerializer):
            phone = PhoneField()


        class AuthorViewSet(ModelViewSet):
            model = ...
            ...

            @SimpleAction(serializer_class=ProfileSerializer, ...)
            def profile(self, request, *args, **kwargs):
                ''' Get profile data to work. '''
                return self.get_object().profile

            @profile.setter
            def profile(self, instance, request, serializer, *args, **kwargs):
                instance.save(update_fields=['phone'])

            @profile.deleter
            def profile(self, instance, request, serializer, *args, **kwargs):
                instance.phone = ''
                instance.save(update_fields=['phone'])
    """
    __slots__ = ('extra_actions',)

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('methods', ['GET'])
        super().__init__(*args, **kwargs)
        self.extra_actions = {}

    def __call__(self, getter=None):
        self.extra_actions['get'] = getter

        def action_method(
                view: viewsets.GenericViewSet,
                request: Request,
                *args, **kwargs,
        ) -> Response:

            get_instance_method = self.extra_actions.get('get')

            if get_instance_method:
                method_kwargs: _t.Dict[_t.Text, _t.Any] = {'query_data': {}}
                if request.method == "GET":
                    try:
                        method_kwargs['query_data'] = view.get_query_serialized_data(request)  # type: ignore
                    except (AttributeError, AssertionError):
                        pass
                instance = get_instance_method(view, request, **method_kwargs)
            else:
                instance = view.get_object() if not self.is_list else view.get_queryset()

            if request.method in {'POST', 'PUT', 'PATCH'}:
                serializer = view.get_serializer(
                    instance,
                    data=request.data,
                    partial=request.method == 'PATCH',
                    many=self.is_list
                )
                serializer.is_valid(raise_exception=True)
            else:
                serializer = view.get_serializer(instance, many=self.is_list)

            if request.method in {'POST', 'PUT', 'PATCH'}:
                serializer.save()
                if 'set' in self.extra_actions:
                    self.extra_actions['set'](view, instance, request, serializer, *args, **kwargs)
                else:
                    instance.save()
            elif request.method == "DELETE":
                if 'del' in self.extra_actions:
                    self.extra_actions['del'](view, instance, request, *args, **kwargs)
                else:
                    instance.delete()

            return serializer.data

        action_method.__name__ = getter.__name__ if getter else self.action_kwargs['name']
        action_method.__doc__ = self.extra_actions['get'].__doc__
        resulted_action = super().__call__(action_method)

        def setter(setter_method):
            self.extra_actions['set'] = setter_method
            if 'PATCH' not in self.methods and 'PUT' not in self.methods:
                self.methods.append('PATCH')
                self.methods.append('PUT')
            setter_method.__name__ = f'{action_method.__name__}__setter'
            return self(self.extra_actions.get('get'))

        def deleter(deleter_method):
            self.extra_actions['del'] = deleter_method
            if 'DELETE' not in self.methods:
                self.methods.append('DELETE')
            deleter_method.__name__ = f'{action_method.__name__}__deleter'
            return self(self.extra_actions.get('get'))

        resulted_action.setter = setter
        resulted_action.deleter = deleter
        resulted_action.__doc__ = self.extra_actions['get'].__doc__
        return resulted_action



from typing import Callable, ClassVar, Iterable, Tuple, Optional
import types
import sys
from functools import partial

from django.conf import settings
try:
    import uwsgi
except ImportError:
    uwsgi = None

from ..utils import BaseVstObject, raise_context_decorator_with_default


METRICS_MAP_TYPE = Iterable[Tuple[Optional[str], Callable]]  # pylint: disable=invalid-name


def get_python_info():
    return {
        'version': f'{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'
    }, 1


def get_metrics_from_workers():  # nocv
    if uwsgi is None:
        return

    workers = uwsgi.workers()

    yield 'uwsgi_total_workers', len(workers)

    for worker in workers:
        worker_data = {
            'id': worker['id'],
            'pid': worker['pid'],
            'status': worker['status'],
        }

        for param in ('requests', 'running_time', 'respawn_count', 'exceptions'):
            yield f'uwsgi_worker_{param}', (worker_data, worker[param])


@raise_context_decorator_with_default(default=0)
def get_uwsgi_metric(metric_name, prep_data=None):  # nocv
    result = uwsgi.metric_get(metric_name)
    if prep_data is not None:
        return prep_data, result
    return result


uwsgi_metrics: list = []
if uwsgi is not None and hasattr(uwsgi, 'metric_get'):  # nocv
    uwsgi_metrics.append(('uwsgi_avg_response_time', partial(get_uwsgi_metric, 'core.avg_response_time')))
    uwsgi_metrics.append(('uwsgi_total_running_time', partial(get_uwsgi_metric, 'core.total_running_time')))
    uwsgi_metrics.append(('uwsgi_idle_workers', partial(get_uwsgi_metric, 'core.idle_workers')))


class BackendMetaClass(type):
    def __new__(mcs, name, bases, attrs, metrics_prefix=None):
        metrics_list = attrs.pop('metrics_list', ())
        if metrics_prefix is not None:
            attrs['prefix'] = metrics_prefix
        result = super().__new__(mcs, name, bases, attrs)
        if metrics_list:
            assert isinstance(metrics_list, Iterable), '"metrics_map" should be iterable instance'
            result._metrics_set = getattr(result, '_metrics_set', ()) + tuple(metrics_list)
        return result


class BaseBackend(
    BaseVstObject,
    metaclass=BackendMetaClass,
    metrics_prefix=settings.VST_PROJECT_LIB
):
    __slots__ = ()
    metrics_list: ClassVar[METRICS_MAP_TYPE] = (
        ('python_info', get_python_info),
        ('{prefix}_database_connections', lambda: (b'', len(settings.DATABASES))),
        ('{prefix}_cache_connections', lambda: (b'', len(settings.CACHES))),
    )
    _metrics_set: ClassVar[METRICS_MAP_TYPE] = ()
    prefix: ClassVar[str] = ''

    def _handle_metrics_data(self, metrics_data: dict):
        for key, value in metrics_data.items():
            if isinstance(value, str):
                yield key, f'"{value}"'
            else:
                yield key, value

    def _handle_metrics_value(self, metrics_name, value):
        if metrics_name is None and isinstance(value, types.GeneratorType):
            for val in value:
                yield from self._handle_metrics_value(*val)
        else:
            if isinstance(value, tuple):
                metrics_data, metrics_value = value
            else:
                metrics_data, metrics_value = b'', value

            yield metrics_name.format(prefix=self.prefix).encode('utf-8')
            if metrics_data:
                if isinstance(metrics_data, dict):
                    yield b'{'
                    yield b','.join(f'{k}={v}'.encode('utf-8') for k, v in self._handle_metrics_data(metrics_data))
                    yield b'}'
                else:
                    yield metrics_data  # nocv
            yield b' '
            if isinstance(metrics_value, (tuple, list)):
                yield ' '.join(map(str, metrics_value)).encode('utf-8')
            else:
                yield str(metrics_value).encode('utf-8')
            yield b'\n'

    def get(self):
        for metrics_name, metrics_method in self._metrics_set:
            yield from self._handle_metrics_value(metrics_name, metrics_method())


class DefaultBackend(BaseBackend):
    __slots__ = ()
    metrics_list = ((None, get_metrics_from_workers), *uwsgi_metrics)



import sys
from subprocess import check_call

from ._base import DockerCommand
from .web import get_celery_command


class Command(DockerCommand):
    help = "Run Celery 'inspect' command"
    with_migration = False

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            'args',
            metavar='celery_arg=value', nargs='*',
            help='Args "name=value" for the celery app.',
        )
        parser.add_argument(
            '-j', '--json',
            action='store_true', dest='json', default=False,
            help="Use json as output format for 'inspect' command.",
        )

    def handle(self, *args, **opts):
        super().handle(*args, **opts)

        cmd_kwargs = dict(
            arg.split('=')
            for arg in args
            if '=' in arg
        )

        cmd_kwargs['_args'] = [
            arg for arg in args
            if '=' not in arg
        ]

        if opts['json']:
            cmd_kwargs['_args'].append('--json')

        cmd = f'{sys.executable} -m {get_celery_command(command="inspect", **cmd_kwargs)}'
        self._print(f'Execute: {cmd}')
        sys.exit(check_call(
            cmd,
            stdout=sys.stdout,
            stderr=sys.stderr,
            stdin=sys.stdin,
            shell=True,  # nosec
        ))



import typing as _t
import sys
import asyncio
import signal
import socket
import time

import uvicorn
import uwsgi  # pylint: disable=import-error
import django
from django.conf import settings
from django.utils.module_loading import import_string
django.setup()


async def init(servers_list, fds):
    config = uvicorn.Config(
        app=import_string(settings.ASGI_APPLICATION),
        interface='asgi3',
        log_level=settings.LOG_LEVEL.lower(),
        **settings.CONFIG['uvicorn'].all(),
    )
    server = uvicorn.Server(config)
    servers_list.append(server)
    await server.serve(sockets=[
        socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        for fd in fds
    ])


def destroy():
    uwsgi.log("destroy worker {0}".format(uwsgi.worker_id()))
    for server in servers:
        server.handle_exit(sig=signal.SIGINT, frame=None)
    time.sleep(1)
    sys.exit(0)


def graceful_reload():
    uwsgi.log("graceful reload for worker {0}".format(uwsgi.worker_id()))
    for server in servers:
        server.handle_exit(sig=signal.SIGTERM, frame=None)
    loop.stop()
    loop.close()
    sys.exit(0)


if __name__ == '__main__':
    # Init event loop and destroy handlers
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, destroy)
    loop.add_signal_handler(signal.SIGTERM, destroy)
    loop.add_signal_handler(signal.SIGHUP, graceful_reload)

    # TODO: Does it nessesary?
    servers: _t.List[uvicorn.Server] = []

    # Spawn a handler for every uWSGI socket
    uwsgi.log(f'start uvicorn worker {uwsgi.worker_id()}')
    loop.run_until_complete(init(servers, uwsgi.sockets))

    # Start accepting requests via socket
    uwsgi.accepting()

    # Graceful shutdown
    destroy()



from typing import Any, Callable, TypeVar, Iterable, Optional, Set, ClassVar, Type, Tuple, Dict, Union, NoReturn, List, Iterator
import abc
from django.db.models import Model, QuerySet, Manager
from django.db.models.manager import RelatedManager
from django.http.response import HttpResponseBase
from rest_framework import views, viewsets, generics, filters, decorators, mixins, serializers, permissions, request as drf_request, response as drf_response
from .base import GenericViewSet

T = TypeVar("T")
_VIEW = TypeVar("_VIEW", bound=Callable[..., HttpResponseBase])
GenericViewSetDecoratorType = Callable[[generics.GenericAPIView], generics.GenericAPIView]
QuerySetFiltersMethodsType = Callable[[Model, QuerySet], QuerySet]
ActionType = Callable[[Union[_VIEW, decorators._View]], decorators.ViewSetAction[_VIEW]]
action = decorators.action
DecoratedActionType = Union[DecoratedActionMixin, decorators.ViewSetAction]


def _create_with_iteration(
        view: Union[GenericViewSet, mixins.CreateModelMixin],
        get_serializer_func: Callable[..., serializers.Serializer],
        nested_manager: RelatedManager,
        data: dict
) -> Model:
    ...


class GenericViewSetInNested(GenericViewSet):
    nested_detail: bool
    nested_allow_append: bool


def get_action_name(master_view: GenericViewSetInNested, method: str = '') -> Optional[str]:
    ...


def subaction(
        methods: Iterable[str] = ('post',),
        detail: bool = None,
        multiaction: bool = False,
        require_confirmation: bool = False,
        is_list: bool = False,
        serializer_class: serializers.Serializer = None,
        response_serializer: serializers.Serializer = None,
        query_serializer: serializers.Serializer = None,
        response_code: int = None,
        permission_classes: Iterable[permissions.BasePermission] = None,
        url_path: str = None,
        description: str = None,
        title: str = None,
        icons: List[str] = None
) -> Callable[[_VIEW], _VIEW]: ...


class DecoratedActionMixin:
    _nested_args: dict
    _nested_manager: Union[str, Callable[[Model], Union[QuerySet, Manager]]]
    _nested_view: generics.GenericAPIView
    _nested_name: str
    _nested_subname: str
    _nested_wrapped_view: 'ApiView'


class ApiView(views.APIView):
    _nested_wrapped_view: 'ApiView'


class NestedViewMixin(viewsets.GenericViewSet):
    nested_manager: Union[QuerySet, Manager]
    queryset_filters: Iterable[QuerySetFiltersMethodsType]
    nested_parent_object: Model
    master_view: GenericViewSetInNested
    nested_allow_append: bool
    nested_id: Union[str, int]
    nested_append_arg: str

    request: drf_request.Request
    check_permissions: Callable[[drf_request.Request], Any]  # type: ignore
    check_throttles: Callable[[drf_request.Request], Any]  # type: ignore
    perform_create: Optional[Callable[[serializers.Serializer], None]]

    def get_serializer_context(self) -> Dict[str, Any]:
        ...

    def finalize_response(self, request: drf_request.Request, response: HttpResponseBase, *args: Any, **kwargs: Any) -> drf_response.Response:
        ...

    def _check_permission_obj(self, objects: Iterable) -> Optional[NoReturn]:
        ...

    def _data_create(self, request_data, nested_append_arg: str) -> List[Union[str, int]]:
        ...

    def create(self, request: drf_request.Request, *_, **__) -> HttpResponseBase:
        ...


class NestedWithoutAppendMixin(NestedViewMixin):
    ...


class NestedWithAppendMixin(NestedWithoutAppendMixin):
    ...


class BaseClassDecorator(metaclass=abc.ABCMeta):
    name: str
    arg: str
    request_arg: str
    args: tuple
    kwargs: dict

    def __init__(self, name: str, arg: Optional[str], *args, **kwargs):
        ...

    @abc.abstractmethod
    def setup(self, view_class: generics.GenericAPIView) -> generics.GenericAPIView:
        ...

    def __call__(self, view_class: generics.GenericAPIView) -> generics.GenericAPIView:
        ...

    def decorator(self, view_class: generics.GenericAPIView) -> generics.GenericAPIView:
        ...


class nested_view(BaseClassDecorator):
    arg: Optional[str]  # type: ignore
    view: GenericViewSet
    allow_append: bool
    append_arg: Optional[str]
    manager_name: Union[str, Callable[[Model], Union[QuerySet, Manager]]]
    methods: Iterable[str]
    queryset_filters: Iterable[QuerySetFiltersMethodsType]
    allowed_subs: Optional[Iterable[str]]
    empty_arg: bool
    _subs: Set[str]

    filter_subs: ClassVar[tuple] = ('filter',)

    NoView: ClassVar[Type[Exception]]

    def __init__(
            self,
            name: str,
            arg: Optional[str],
            view: GenericViewSet,
            allow_append: bool = False,
            manager_name: Union[str, Callable[[Model], Union[QuerySet, Manager]]] = '',
            queryset_filters: Iterable[QuerySetFiltersMethodsType] = (),
            methods: Iterable[str] = (),
            subs: Optional[Iterable[str]] = None,
            append_arg: Optional[str] = None,
            empty_arg: bool = False,
            *args,
            **kwargs
    ): ...

    def _get_subs_from_view(self) -> Iterable[str]:
        ...

    def get_subs(self) -> Set[str]:
        ...

    def get_view(self, name: str, **options) -> Tuple[str, ApiView]:
        ...

    def get_view_type(self, type_name: str, **options) -> Tuple[str, ApiView]:
        ...

    def get_list_view(self, **options) -> Tuple[str, ApiView]:
        ...

    def get_detail_view(self, **options) -> Tuple[str, ApiView]:
        ...

    def get_sub_view(self, sub, **options) -> Tuple[str, ApiView]:
        ...

    def get_decorator(self, detail: bool = False, **options) -> ActionType:
        ...

    def _filter_methods(self, methods: Iterable[str], detail: bool = False) -> Set[str]:
        ...

    def decorated(self, detail: bool) -> Tuple[str, DecoratedActionType]:
        ...

    def _get_decorated_sub(self, sub: str) -> Tuple[str, DecoratedActionType]:
        ...

    def decorated_list(self):
        return self.decorated(detail=False)

    def decorated_detail(self):
        return self.decorated(detail=True)

    def generate_decorated_subs(self) -> Iterator[Tuple[str, DecoratedActionType]]:
        ...

    def setup(self, view_class: generics.GenericAPIView) -> generics.GenericAPIView:
        ...


def extend_viewset_attribute(name, override=False, data=None) -> GenericViewSetDecoratorType:
    ...

def extend_filterbackends(backends: Iterable[filters.BaseFilterBackend], override: bool = False) -> GenericViewSetDecoratorType:
    ...

def cache_method_result(func: Callable[..., T]) -> Callable[..., T]:
    ...



import os
import posixpath

from django.conf import settings
from django.core.asgi import get_asgi_application
from django.contrib.staticfiles import finders
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse, FileResponse
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.staticfiles import StaticFiles


NOT_FOUND_RESPONSE = PlainTextResponse('Not found', status_code=404)
static_app = FastAPI(openapi_url=None, docs_url=None, redoc_url=None)
static_app.add_middleware(GZipMiddleware)

application = FastAPI(openapi_url=None, docs_url=None, redoc_url=None)
application.mount(settings.STATIC_URL, static_app)

if settings.HAS_DOCS:
    application.mount(
        settings.DOC_URL,
        GZipMiddleware(StaticFiles(directory=settings.DOCS_ROOT, html=True)),
        name="docs"
    )

if os.path.exists(settings.MEDIA_ROOT) and settings.MEDIA_URL:
    application.mount(
        settings.MEDIA_URL,
        GZipMiddleware(StaticFiles(directory=settings.MEDIA_ROOT)),
        name="media"
    )


@static_app.get('/{file_path:path}')
async def static(file_path: str):
    absolute_path = finders.find(posixpath.normpath(file_path).lstrip("/"))
    if not absolute_path or os.path.isdir(absolute_path):
        return NOT_FOUND_RESPONSE
    return FileResponse(absolute_path)

application.mount("/", get_asgi_application())



# pylint: disable=invalid-name
import os
import re
import functools

from django.conf import settings
from django.urls.conf import include, re_path, path
from django.contrib import admin
from django.views.static import serve
from django.contrib.staticfiles.views import serve as serve_static
from django.views.generic.base import RedirectView
from rest_framework import permissions

from .api.routers import MainRouter
from .utils import URLHandlers
from .api.views import HealthView, MetricsView

static_serve = functools.partial(serve_static, insecure=True)


def get_valid_url(*args):
    return os.path.join('/', *map(lambda x: x.lstrip('^/'), filter(bool, args + ('/',))))


class AdminLoginLogoutRedirectView(RedirectView):
    query_string = True


# Main router for all APIs versions
router = MainRouter(perms=(permissions.IsAuthenticated,))
router.generate_routers(settings.API)
router.register_view('health', HealthView.as_view({'get': 'list'}), 'health')
if settings.ENABLE_METRICS_VIEW:
    router.register_view('metrics', MetricsView.as_view({'get': 'list'}), 'metrics')


admin.site.site_header = 'Admin panel'
admin.site.site_title = settings.VST_PROJECT
admin.site.index_title = f"{settings.VST_PROJECT.upper()} Settings Panel"
admin.site.site_url = "/"
admin.site.login = AdminLoginLogoutRedirectView.as_view(  # type: ignore
    url=get_valid_url(settings.ACCOUNT_URL, settings.LOGIN_URL)
)
admin.site.logout = AdminLoginLogoutRedirectView.as_view(  # type: ignore
    url=get_valid_url(settings.ACCOUNT_URL, settings.LOGOUT_URL)
)
doc_url = getattr(settings, 'DOC_URL', '/docs/')[1:]

urlpatterns = list(URLHandlers())

urlpatterns += [
    path(settings.ACCOUNT_URL, include(list(URLHandlers('ACCOUNT_URLS'))))
]

urlpatterns += [
    path('admin/', admin.site.urls),
] if getattr(settings, 'ENABLE_ADMIN_PANEL', False) else []

urlpatterns += [re_path(rf'^{settings.API_URL}/', include(router.urls))]
if settings.STATIC_URL.startswith('/'):
    urlpatterns.append(
        re_path(r"^%s(?P<path>.*)$" % re.escape(settings.STATIC_URL.lstrip("/")), static_serve)
    )
if settings.MEDIA_URL.startswith('/') and settings.MEDIA_ROOT:
    urlpatterns.append(re_path(
        r"^%s(?P<path>.*)$" % re.escape(settings.MEDIA_URL.lstrip("/")),
        functools.partial(serve, document_root=settings.MEDIA_ROOT)
    ))
if settings.HAS_DOCS:
    urlpatterns +=  [
        path(
            f'{settings.DOC_URL[1:]}',
            view=serve,
            kwargs={'path': 'index.html', 'document_root': settings.DOCS_ROOT}
        ),
        re_path(
            r"^%s(?P<path>.*)$" % re.escape(settings.DOC_URL.lstrip("/")),
            functools.partial(serve, document_root=settings.DOCS_ROOT)
        )
    ]



# pylint: disable=abstract-method,unused-import,undefined-variable
import json
import os
import logging
import tempfile
import sys
import time
import traceback
from pathlib import Path
import subprocess
from collections import OrderedDict

from environ import Env
from django.conf import settings
from django.core.management.base import (
    BaseCommand as _BaseCommand,
    CommandError as CmdError,
)
from configparserc.config import ConfigParserC, Section, BoolType
from vstutils.utils import Lock


tmp = tempfile.tempdir
logger = logging.getLogger(settings.VST_PROJECT)
logger_lib = logging.getLogger(settings.VST_PROJECT_LIB)


class BaseCommand(_BaseCommand):
    interactive = False
    requires_system_checks = ()
    keep_base_opts = False
    help = "Service command for web-application"

    class CommandError(CmdError):
        pass

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.fromfile_prefix_chars = '@'
        parser.add_argument(
            '-l', '--log-level',
            action='store',
            dest='log-level',
            default=False,
            type=str,
            help='Set logs level [debug|warning|error|critical]')
        if self.interactive:
            parser.add_argument(
                '--noinput', '--no-input',
                action='store_false', dest='interactive', default=True,
                help="Do NOT prompt the user for input of any kind.",
            )

    def _print(self, msg, style=None):
        style = style or 'HTTP_INFO'
        style = getattr(self.style, style, str)
        self.stdout.write(style(msg))

    def _settings(self, value, default=None):
        return getattr(settings, value, default)

    def _get_versions(self):
        versions = OrderedDict(django=super().get_version())
        versions[self._settings('VST_PROJECT')] = self._settings('PROJECT_VERSION')
        if self._settings('VST_PROJECT') != self._settings('VST_PROJECT_LIB'):
            versions[self._settings('VST_PROJECT_LIB')] = self._settings(
                'PROJECT_LIB_VERSION'
            )
        if self._settings('VST_PROJECT') != "vstutils":
            versions['vstutils'] = self._settings('VSTUTILS_VERSION')
        return versions

    def get_version(self):
        return ' '.join([
            f'{k.title()}={v}' for k, v in self._get_versions().items()
        ])

    def ask_user(self, message, default=None):
        # pylint: disable=import-error
        if getattr(self, 'interactive_mode', False):
            return input(message) or default
        return default

    def ask_user_bool(self, message, default=True):
        reply = self.ask_user(message, 'yes' if default else 'no').lower()
        if reply in ['y', 'yes']:
            return True
        elif reply in ['n', 'no']:
            return False

    def handle(self, *args, **options):
        # pylint: disable=invalid-name
        self.interactive_mode = options.pop('interactive', False)
        LOG_LEVEL = settings.LOG_LEVEL
        if options.get('log-level', False):
            LOG_LEVEL = options.pop('log-level', LOG_LEVEL)
        logger.setLevel(LOG_LEVEL.upper())
        logger_lib.setLevel(LOG_LEVEL.upper())
        self.LOG_LEVEL = LOG_LEVEL.upper()
        os.environ.setdefault('DJANGO_LOG_LEVEL', self.LOG_LEVEL)


class DockerCommand(BaseCommand):
    interactive = True
    with_migration = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        if self.with_migration:
            parser.add_argument(
                '--migrate-attempts', '-a',
                default=60,
                dest='attempts', help='The number of attempts to migrate.',
            )
            parser.add_argument(
                '--migrate-attempts-sleep-time', '-t',
                default=1,
                dest='attempts_timeout', help='The number of attempts to migrate.',
            )

    def handle(self, *args, **options):
        super().handle(*args, **options)
        self.prefix = self._settings('VST_PROJECT_LIB', 'vstutils').upper()
        self.project_name = self._settings('VST_PROJECT', 'vstutils')

        logger.debug(f'Prefix={self.prefix} | Project={self.project_name}')

        self.env = os.environ.copy()
        config = self.prepare_config()
        self.env[self._settings('CONFIG_ENV_DATA_NAME')] = config.generate_config_string()
        default_envs = {
            'UWSGI_PROCESSES': 'UWSGI_WORKERS',
            'UWSGI_THREADS': 'UWSGI_THREADS'
        }
        for key in default_envs:  # pylint: disable=consider-using-dict-items
            value = os.environ.get(f"{self.prefix}_{key}", '')
            if value:
                self.env[default_envs[key]] = value  # nocv

        if config['main']['debug'] or self._settings('TESTS_RUN', False):
            logger.debug(f'Env:\n{json.dumps(self.env, indent=4)}')
            logger.debug(f'Config:\n{self.env[self._settings("CONFIG_ENV_DATA_NAME")]}')

    @property
    def databases_to_migrate(self):
        allowed_databases = ('default',) + tuple(set(getattr(settings, 'DOCKER_DATABASES_TO_MIGRATE', ())))
        for db in settings.DATABASES.keys():
            if db in allowed_databases:
                yield db

    def migrate(self, options, *args):
        success = False
        error = 'Unknown error.'
        for i in range(options.get('attempts', 60)):
            try:
                with Lock(
                        self._settings('DOCKERRUN_MIGRATE_LOCK_ID'),
                        1,
                        self._settings('DOCKERRUN_MIGRATE_LOCK_TIMEOUT'),
                        'Migration process still locked by another application process.',
                        int(os.getenv("DOCKER_MIGRATE_LOCK_KEY_TIMEOUT", '0')) or None
                ) as lock:
                    logger.info(f'Migration locked by key: `{lock.id}`')
                    for db_name in self.databases_to_migrate:
                        logger.info(f'Migrating db "{db_name}".')
                        subprocess.check_call(
                            [
                                sys.executable,
                                '-m',
                                self.project_name,
                                'migrate',
                                '--database',
                                db_name,
                                *args
                            ],
                            env=self.env,
                            bufsize=0,
                            universal_newlines=True,
                        )
                    logger.info(f'Unlocking migration by key: `{lock.id}`')
            except:
                error = traceback.format_exc()
                logger.debug(f'Migration attempt {i} failed: {sys.exc_info()[1]}')
                self._print(f"Retry #{i}...", 'WARNING')
                time.sleep(options.get('attempts_timeout', 1))
            else:
                success = True
                break
        return success, error

    def prepare_section_main(self, config):
        config['main'] = {
            'debug': os.getenv(f'{self.prefix}_DEBUG', self._settings('DEBUG')),
            'log_level': self.log_level,
            'timezone': os.getenv(f'{self.prefix}_TIMEZONE', self._settings('TIME_ZONE')),
            'enable_admin_panel': os.getenv(f'{self.prefix}_ENABLE_ADMIN_PANEL', 'false'),
            'first_day_of_week': os.getenv(
                f'{self.prefix}_FIRST_DAY_OF_WEEK',
                self._settings('FIRST_DAY_OF_WEEK')
            )
        }
        # ldap-server, ldap-default-domain if exist
        ldap_server = os.getenv(f'{self.prefix}_LDAP_CONNECTION', None)
        if ldap_server:  # nocv
            config['main']['ldap_server'] = ldap_server
        ldap_default_domain = os.getenv(f'{self.prefix}_LDAP_DOMAIN', None)
        if ldap_default_domain:  # nocv
            config['main']['ldap-default-domain'] = ldap_default_domain

    def prepare_section_db(self, config):
        # Check if configured via default url
        if Env.DEFAULT_DATABASE_ENV in os.environ:  # nocv
            return  # Get data via standard settings mechanism

        # SQLite prepearing
        sqlite_default_dir = os.path.join(os.environ.get(f'{self.prefix}_SQLITE_DIR', str(Path.home())), '')
        if not os.path.exists(sqlite_default_dir):  # nocv
            os.makedirs(sqlite_default_dir)
        sqlite_db_path = os.path.join(sqlite_default_dir, os.environ.get(f'{self.prefix}_SQLITE_DBNAME', 'db.sqlite3'))

        if os.getenv(f'{self.prefix}_DB_HOST') is not None:
            try:
                pm_type = os.getenv(f'{self.prefix}_DB_TYPE', 'mysql')

                default_port = ''
                if pm_type == 'mysql':
                    default_port = '3306'
                elif pm_type == 'postgresql':  # nocv
                    default_port = '5432'

                config['database'] = {
                    'engine': f'django.db.backends.{pm_type}',
                    'name': os.environ[f'{self.prefix}_DB_NAME'],
                    'user': os.environ[f'{self.prefix}_DB_USER'],
                    'password': os.environ[f'{self.prefix}_DB_PASSWORD'],
                    'host': os.environ[f'{self.prefix}_DB_HOST'],
                    'port': os.getenv(f'{self.prefix}_DB_PORT', default_port),
                }
                config['database.options'] = {
                    'connect_timeout': os.getenv(f'{self.prefix}_DB_CONNECT_TIMEOUT', '20'),
                }
                if pm_type == 'mysql':
                    config['database.options']['init_command'] = os.getenv('DB_INIT_CMD', '')
            except KeyError as err:  # nocv
                raise Exception(
                    f'Not enough variables for connect to  SQL server. Variable {str(err)} is required.'
                ) from err
        else:  # nocv
            config['database'] = {
                'engine': 'django.db.backends.sqlite3',
                'name': sqlite_db_path
            }

    def prepare_section_cache(self, config):
        # Check if configured via default url
        if Env.DEFAULT_CACHE_ENV in os.environ:  # nocv
            return  # Get data via standard settings mechanism

        cache_type = os.getenv(f'{self.prefix}_CACHE_TYPE', 'file')
        if cache_type == 'file':
            cache_engine = 'django.core.cache.backends.filebased.FileBasedCache'
        elif cache_type == 'memcache':  # nocv
            cache_engine = 'django.core.cache.backends.memcached.MemcachedCache'
        elif cache_type == 'redis':  # nocv
            cache_engine = 'django.core.cache.backends.redis.RedisCache'
        else:  # nocv
            raise Exception(f'Unknown cache type `{cache_type}`.')

        config['cache'] = config['locks'] = config['session'] = config['etag'] = {
            'backend': cache_engine,
            'location': os.getenv('CACHE_LOCATION', f'{tmp}/{self.prefix}_django_cache')
        }

    def prepare_section_rpc(self, config):
        rpc_connection = os.getenv('RPC_ENGINE', None)
        config['rpc'] = {
            'heartbeat': os.getenv('RPC_HEARTBEAT', '5'),
            'concurrency': os.getenv('RPC_CONCURRENCY', '4'),
            'clone_retry_count': os.getenv('RPC_CLONE_RETRY_COUNT', '3')
        }

        if rpc_connection:  # nocv
            config['rpc']['connection'] = rpc_connection
        config['rpc']['enable_worker'] = 'false'

        # Set worker settings
        if os.environ.get('WORKER', '') == 'ENABLE':  # nocv
            config['rpc']['enable_worker'] = 'true'
            config['worker'] = {
                'loglevel': self.log_level,
                'pidfile': f'{tmp}/{self.prefix.lower()}_worker.pid',
                'beat': os.getenv(f'{self.prefix}_SCHEDULER_ENABLE', 'true')
            }

    def prepare_section_web(self, config):
        config['web'] = {
            'session_timeout': os.getenv(f'{self.prefix}_SESSION_TIMEOUT', '2w'),
            'rest_page_limit': os.getenv(f'{self.prefix}_WEB_REST_PAGE_LIMIT', '100'),
            'enable_gravatar': os.getenv(f'{self.prefix}_WEB_GRAVATAR', 'true'),
            'request_max_size': os.getenv(
                f'{self.prefix}_REQUEST_MAX_SIZE',
                self._settings('DATA_UPLOAD_MAX_MEMORY_SIZE')
            ),
            'x_frame_options': os.getenv(
                f'{self.prefix}_X_FRAME_OPTIONS',
                self._settings('X_FRAME_OPTIONS')
            ),
            'use_x_forwarded_host': os.getenv(
                f'{self.prefix}_USE_X_FORWARDED_HOST',
                self._settings('USE_X_FORWARDED_HOST')
            ),
            'use_x_forwarded_port': os.getenv(
                f'{self.prefix}_USE_X_FORWARDED_PORT',
                self._settings('USE_X_FORWARDED_PORT')
            ),
            'password_reset_timeout_days': os.getenv(
                f'{self.prefix}_PASSWORD_RESET_TIMEOUT_DAYS',
                self._settings('PASSWORD_RESET_TIMEOUT_DAYS')
            ),
            'secure_browser_xss_filter': os.getenv(
                f'{self.prefix}_SECURE_BROWSER_XSS_FILTER',
                self._settings('SECURE_BROWSER_XSS_FILTER')
            ),
            'secure_content_type_nosniff': os.getenv(
                f'{self.prefix}_SECURE_CONTENT_TYPE_NOSNIFF',
                self._settings('SECURE_CONTENT_TYPE_NOSNIFF')
            ),
            'secure_hsts_include_subdomains': os.getenv(
                f'{self.prefix}_SECURE_HSTS_INCLUDE_SUBDOMAINS',
                self._settings('SECURE_HSTS_INCLUDE_SUBDOMAINS')
            ),
            'secure_hsts_preload': os.getenv(
                f'{self.prefix}_SECURE_HSTS_PRELOAD',
                self._settings('SECURE_HSTS_PRELOAD')
            ),
            'secure_hsts_seconds': os.getenv(
                f'{self.prefix}_SECURE_HSTS_SECONDS',
                self._settings('SECURE_HSTS_SECONDS')
            ),
            'health_throttle_rate': os.getenv(
                f'{self.prefix}_HEALTH_THROTTLE_RATE',
                self._settings('HEALTH_THROTTLE_RATE').replace('/minute', '')
            ),
        }

    def prepare_section_uwsgi(self, config):
        config['uwsgi'] = {
            'thread-stacksize': os.getenv(f'{self.prefix}_UWSGI_THREADSTACK', '40960'),
            'max-requests': os.getenv(f'{self.prefix}_UWSGI_MAXREQUESTS', '50000'),
            'limit-as': os.getenv(f'{self.prefix}_UWSGI_LIMITS', '1024'),
            'pidfile': os.getenv(f'{self.prefix}_UWSGI_PIDFILE', f'{tmp}/{self.prefix.lower()}_web.pid'),
            'daemon': 'false'
        }

        current_addr, current_port = self._settings('WEB_ADDRPORT').split(',')[0].split(':')
        config['uwsgi']['addrport'] = (
            f"{os.getenv(f'{self.prefix}_WEB_HOST', current_addr)}:"
            f"{os.getenv(f'{self.prefix}_WEB_PORT', current_port)}"
        )
        harakiri = os.getenv(f'{self.prefix}_UWSGI_HARAKIRI', '')
        if harakiri:
            config['uwsgi']['harakiri'] = harakiri
        vacuum = os.getenv(f'{self.prefix}_UWSGI_VACUUM', '')
        if vacuum:
            config['uwsgi']['vacuum'] = vacuum

    def prepare_section_smtp(self, config):
        mail_parameters = ['port', 'user', 'password', 'tls', 'ssl', 'from_address']
        mail_settings = {'host': os.environ.get(f'{self.prefix}_MAIL_HOST')}
        if mail_settings['host']:
            for param in mail_parameters:
                value = os.environ.get(f'{self.prefix}_MAIL_{param.upper()}')
                if value:
                    mail_settings[param] = value
            config['mail'] = mail_settings

    def prepare_config(self):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        prefix = self.prefix

        default_sections = (
            'main',
            'db',
            'cache',
            'rpc',
            'web',
            'uwsgi',
            'smtp',
        )
        section_handlers_prefix = 'prepare_section_'
        sections = default_sections + tuple(map(
            lambda y: y.replace(section_handlers_prefix, ''),
            filter(
                lambda x: x.startswith(section_handlers_prefix) and x not in default_sections,
                dir(self)
            )
        ))

        class DockerSection(Section):
            types_map = {
                f'override_{s}': BoolType()
                for s in sections
            }

        # Start configuring config file
        self.config = ConfigParserC(
            format_kwargs=self._settings('KWARGS', {}),
            section_defaults={
                'docker': {
                    s: os.getenv(f'{prefix}_DOCKER_{s.upper()}', 'True')
                    for s in DockerSection.types_map
                }
            },
            section_overload={
                'docker': DockerSection
            },
            format_exclude_sections=('uwsgi',)
        )
        config = self.config
        config.parse_text(settings.CONFIG.generate_config_string())

        # Set log level
        self.log_level = os.getenv(f'{prefix}_LOG_LEVEL', 'WARNING')

        for section_name in sections:
            if config['docker'].getboolean(f'override_{section_name}', fallback=True):
                getattr(self, f'prepare_section_{section_name}')(config)

        # Set secret key
        os.environ.setdefault('SECRET_KEY', 'DISABLE')
        if os.environ['SECRET_KEY'] != 'DISABLE':  # nocv
            with open(f'/etc/{prefix.lower()}/secret', 'w', encoding='utf-8') as secretfile:
                secretfile.write(os.environ['SECRET_KEY'])

        return config



# cython: binding=True
import typing as _t
import logging
import json

import orjson
from django.db.models import signals
from django.conf import settings
from django.contrib.auth import get_user_model
from cent import Client as CentrifugoClient  # type: ignore

from .base import get_proxy_labels
from ..utils import raise_context_decorator_with_default, raise_context
from .model import BaseModel
from ..api.renderers import ORJSONRenderer

logger = logging.getLogger('vstutils')


class JsonEncoder(json.JSONEncoder):
    def default(self, o):
        return str(o)  # nocv


class Notificator:
    client_class = CentrifugoClient

    queue: _t.List[_t.Tuple[_t.Sequence[_t.Text], _t.Any]]
    cent_client: CentrifugoClient
    label: _t.Text

    _json_renderer = ORJSONRenderer()

    def __init__(self, queue=None, client=None, label=None):
        self.queue = queue or []
        self.cent_client = client or self.get_client()
        self.label = label
        self._signals: _t.List[signals.ModelSignal] = []
        if self.cent_client is not None:
            self.connect_signal(signals.post_save)
            self.connect_signal(signals.post_delete)

    def connect_signal(self, signal: signals.ModelSignal):
        if signal not in self._signals:
            signal.connect(self.signal_handler)
            self._signals.append(signal)

    def disconnect_signal(self, signal: signals.ModelSignal):
        if signal in self._signals:
            signal.disconnect(self.signal_handler)

    def signal_handler(self, instance, *args, **kwargs):
        if isinstance(instance, (BaseModel, get_user_model())) and getattr(instance, '_notify_update', True):
            self.create_notification_from_instance(instance)

    def get_openapi_secret(self):
        return settings.CENTRIFUGO_CLIENT_KWARGS.get('token_hmac_secret_key', '')

    @raise_context_decorator_with_default(verbose=False)
    def get_client(self):
        centrifugo_client_kwargs = {**settings.CENTRIFUGO_CLIENT_KWARGS}
        centrifugo_client_kwargs.pop('token_hmac_secret_key', None)
        if 'json_encoder' not in centrifugo_client_kwargs:
            centrifugo_client_kwargs['json_encoder'] = JsonEncoder
        logger.debug(f"Getting Centrifugo client with kwargs: {centrifugo_client_kwargs}")
        return self.client_class(**centrifugo_client_kwargs)

    def create_notification_from_instance(self, instance):  # pylint: disable=invalid-name
        model = instance.__class__
        self.queue.append(
            ((model._meta.label, *get_proxy_labels(model)), {'pk': instance.pk})
        )

    def create_notification(self, labels, data):
        if isinstance(labels, str):
            labels = (labels,)
        data = orjson.loads(self._json_renderer.render(data) or '{}')
        self.queue.append(
            (labels, data)
        )

    def clear_messages(self):
        self.queue.clear()

    @raise_context_decorator_with_default()
    def send(self):
        self.queue, objects = [], tuple(self.queue)

        sent_channels = set()
        provided_label = self.label
        for obj_labels, data in objects:
            with raise_context():
                for obj_label in obj_labels:
                    channel = self.get_subscription_channel(provided_label or obj_label)
                    self.cent_client.add("publish", self.cent_client.get_publish_params(
                        channel=channel,
                        data=data,
                    ))
                    sent_channels.add(channel)
        if objects:
            logger.debug(f'Send notifications about {len(objects)} updates to channel(s) {sent_channels}.')
            return self.cent_client.send()

    def disconnect_all(self):
        for signal in self._signals:
            self.disconnect_signal(signal)

    def get_subscription_channel(self, label: str):
        return f'{settings.SUBSCRIPTIONS_PREFIX}.{label}'

    def __del__(self):
        self.disconnect_all()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.send()
        self.disconnect_all()



import json
import sys

from celery.result import AsyncResult
from django.utils.module_loading import import_string

from ._base import DockerCommand


class Command(DockerCommand):
    interactive = True
    help = "Run tasks in container environment, like a CronJob."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--sync',
            action='store_true',
            dest='sync',
            default=False,
            help="Run task in sync mode.",
        )
        parser.add_argument(
            '-k', '--kwargs',
            action='store',
            dest='kwargs',
            default='{}',
            type=str,
            help='Setup named arguments as json.'
        )
        parser.add_argument(
            '--task',
            action='store',
            dest='task',
            default='',
            type=str,
            help='Setup named arguments as json.'
        )

    def _parse_arg(self, func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            self._print(str(err), "ERROR")
            sys.exit(-1)

    def handle(self, *args, **options):
        super().handle(*args, **options)

        task = options['task']
        self._print(f'Run "{task}"', 'WARNING')

        kwargs, task = self._parse_arg(json.loads, options['kwargs']), self._parse_arg(import_string, options['task'])()
        task_result: AsyncResult = self._parse_arg(task.apply if options['sync'] else task.apply_async, kwargs=kwargs)

        self._print(f'Task "{task}" is executed with id "{task_result.task_id}".', 'SUCCESS')
